/***************************************************************************
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <math.h>
#include "mptkcodec.h"

mdct::mdct( unsigned int set_len, unsigned int set_windowType, double set_windowOpt ) {

	extern MP_Win_Server_c MP_GLOBAL_WIN_SERVER;

	/* Set parameters */
	len = set_len;
	numFreqs = len >> 1;
	windowType = set_windowType;
	windowOpt = set_windowOpt;

	/* Init fft */
	infft = (fftw_complex*) fftw_malloc( sizeof(fftw_complex) * len );
	outfft = (fftw_complex*) fftw_malloc( sizeof(fftw_complex) * len );
	p = fftw_plan_dft_1d( (int)(len), infft, outfft, FFTW_FORWARD, FFTW_ESTIMATE );
	infftInv = (fftw_complex*) fftw_malloc( sizeof(fftw_complex) * len );
	outfftInv = (fftw_complex*) fftw_malloc( sizeof(fftw_complex) * len );
	iP = fftw_plan_dft_1d( (int)(len), infftInv, outfftInv, FFTW_BACKWARD, FFTW_ESTIMATE );

	/* Init arrays */
	samples = new double[len];
	mcltRe = new double[numFreqs];
	mcltIm = new double[numFreqs];
	mcltE = new double[numFreqs];
	preModRe = new double[len];
	preModIm = new double[len];
	postModRe = new double[numFreqs];
	postModIm = new double[numFreqs];
	preModReInv = new double[numFreqs];
	preModImInv = new double[numFreqs];
	postModReInv = new double[len];
	postModImInv = new double[len];

	/* Init window */
	MP_GLOBAL_WIN_SERVER.get_window( &window, len, windowType, windowOpt );

	/* Init transform */
	unsigned int i;
	for ( i = 0 ; i < len ; i++ ) {
		*(preModRe+i) = cos( MP_PI * i / len );
		*(preModIm+i) = - sin( MP_PI * i / len );
	}
	for ( i = 0 ; i < numFreqs ; i++ ) {
		*(postModRe+i) = cos( MP_PI * (1 + len*0.5 ) * ( i + 0.5 ) / len );
		*(postModIm+i) = - sin( MP_PI * (1 + len*0.5 ) * ( i + 0.5 ) / len );
	}
	for ( i = 0 ; i < numFreqs; i++ ) {
		*(preModReInv+i) = cos( 2 * MP_PI * i * ( 0.5 + len*0.25 ) / len );
		*(preModImInv+i) = sin( 2 * MP_PI * i * ( 0.5 + len*0.25 ) / len );
	}
	for ( i = 0 ; i < len ; i++ ) {
		*(postModReInv+i) = cos( MP_PI * ( i + 0.5 + len*0.25 ) / len );
		*(postModImInv+i) = sin( MP_PI * ( i + 0.5 + len*0.25 ) / len );
	}
	for ( i=0; i<len; i++ ) {
		infftInv[i][0] = 0;
		infftInv[i][1] = 0;
	}
}

mdct::~mdct( void ) {

	/* Clean memory */
	fftw_destroy_plan( p );
	fftw_destroy_plan( iP );
	fftw_free( infft );
	fftw_free( outfft );
	fftw_free( infftInv );
	fftw_free( outfftInv );
	delete[]( samples );
	delete[]( mcltRe );
	delete[]( mcltIm );
	delete[]( mcltE );
	delete[]( preModRe );
	delete[]( preModIm );
	delete[]( postModRe );
	delete[]( postModIm );
	delete[]( preModReInv );
	delete[]( preModImInv );
	delete[]( postModReInv );
	delete[]( postModImInv );
}

void mdct::compute_mclt( double *in ) {

	unsigned int i;

	/* Modulate and window the input signal */
	for ( i=0; i<len; i++ ) {
		infft[i][0] = preModRe[i] * window[i] * in[i];
		infft[i][1] = preModIm[i] * window[i] * in[i];
	}

	/* Compute fast fourier transform */
	fftw_execute( p );

	/* Modulate the fft output */
	for ( i = 0 ; i < numFreqs ; i++ ) {
		mcltRe[i] = 2 * outfft[i][0] * postModRe[i] - 2 * outfft[i][1] * postModIm[i];
		mcltIm[i] = 2 * outfft[i][1] * postModRe[i] + 2 * outfft[i][0] * postModIm[i];
		mcltE[i] = mcltRe[i]*mcltRe[i] + mcltIm[i]*mcltIm[i];
	}
}

void mdct::compute_inv_mdct( double *in ) {

	unsigned int i;

	/* Modulate the input coefficients */
	for ( i=0; i<numFreqs; i++ ) {
		infftInv[i][0] = preModReInv[i] * in[i];
		infftInv[i][1] = preModImInv[i] * in[i];
	}

	/* Inverse fast fourier transform */
	fftw_execute( iP );

	/* Modulate and window the ifft output */
	for ( i=0; i<len; i++ ) {
		samples[i] = postModReInv[i] * window[i] * outfftInv[i][0] - postModImInv[i] * window[i] * outfftInv[i][1];
	}
}
