/***************************************************************************
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <math.h>
#include "mptkcodec.h"

slot::slot(unsigned int set_numBlocks, dictionary *set_dict) {

	/* Declarations */
	unsigned int l,n;

	/* Set parameters */
	dict = set_dict;
	numBlocks = set_numBlocks;

	/* Init arrays */
	numFrames = new unsigned int[numBlocks];
	frameShift = new int[numBlocks];
	numFreqs = new unsigned int[numBlocks];
	data = new double**[numBlocks];
	inter = new unsigned int**[numBlocks];
	for ( l=0 ; l<numBlocks ; l++ ) {
		numFrames[l] = dict[numBlocks-1].windowLen / dict[l].windowLen;
		numFreqs[l] = dict[l].windowLen >> 1;
		frameShift[l] = - (numFreqs[l] >> 1);
		data[l] = new double*[numFrames[l]];
		inter[l] = new unsigned int*[numFrames[l]];
		for ( n=0 ; n<numFrames[l] ; n++ ) {
			data[l][n] = new double[numFreqs[l]];
			inter[l][n] = new unsigned int[numFreqs[l]];
		}
	}
	dataI_len = numFreqs[numBlocks-1]*numBlocks;
	dataI = new double[dataI_len];
	signal_len = numFreqs[numBlocks-1]*2;
	signal = new double[signal_len];

	/* Init interleaving matrix */
	init_interleaving();

	/* Init other variables */
	mdct_blocks = NULL;
	psy_blocks = NULL;
}

slot::~slot( void ) {

	/* Declarations */
	unsigned int l, n;

	/* Delete */
	delete[]( frameShift );
	delete[]( numFreqs );
	for ( l=0 ; l<numBlocks ; l++ ) {
		for ( n=0 ; n<numFrames[l] ; n++ ) {
			delete[]( data[l][n] );
			delete[]( inter[l][n] );
		}
		delete[]( data[l] );
		delete[]( inter[l] );
	}
	delete[]( numFrames );
	delete[]( data );
	delete[]( inter );
	delete[]( dataI );
	delete[]( signal );
	if ( mdct_blocks ) {
		for ( l=0 ; l<numBlocks ; l++ ) {
			delete( mdct_blocks[l] );
		}
		delete[]( mdct_blocks );
	}
	if ( psy_blocks ) {
		for ( l=0 ; l<numBlocks ; l++ ) {
			delete( psy_blocks[l] );
		}
		delete[]( psy_blocks );
	}
}

void slot::init_mdct( void ) {

	/* Declarations */
	unsigned int l;

	/* Init MDCT for each block */
	mdct_blocks = new mdct*[numBlocks];
	for ( l=0 ; l<numBlocks ; l++ ) {
		mdct_blocks[l] = new mdct(dict[l].windowLen,dict[l].windowType,dict[l].windowOption);
	}
}

void slot::init_psy( unsigned int sampleRate ) {

	/* Declarations */
	unsigned int l;

	/* Init Psy-model for each block */
	psy_blocks = new psy*[numBlocks];
	for ( l=0 ; l<numBlocks ; l++ ) {
		psy_blocks[l] = new psy(dict[l].windowLen>>1, sampleRate );
	}
}

void slot::load_atoms( MP_Book_c *book, unsigned int slotPos ) {

	/* Declarations */
	unsigned int n;
	unsigned int block;
	unsigned int frame;
	unsigned int freq;

	/* Fill array with zeros */
	reset_data();

	/* Add atoms amplitude */
	for ( n=0 ; n<book->numAtoms ; n++ ) {
		block = (unsigned int) round( log2(book->atom[n]->support[0].len) - log2(numFreqs[0]) - 1 );
		frame = (book->atom[n]->support[0].pos - frameShift[block] - slotPos) / numFreqs[block];
		freq = (unsigned int) ( (book->atom[n]->get_field(MP_FREQ_PROP,0) * numFreqs[block] * 2) - 0.5 );
		data[block][frame][freq] += book->atom[n]->amp[0];
	}
}

void slot::save_atoms( MP_Book_c *book, unsigned int slotPos ) {

	/* Declarations */
	unsigned int l,n,m;
	MP_Mdct_Atom_c *atom;

	/* Deinterleave data */
	deinterleave_data();

	/* Find non-zero coefficients */
	for ( l=0 ; l<numBlocks ; l++ ) {
		for ( m=0 ; m<numFrames[l] ; m++ ) {
			for ( n=0 ; n<numFreqs[l] ; n++ ) {
				if (data[l][m][n] != 0) {
					atom = MP_Mdct_Atom_c::init( 1, dict[l].windowType, dict[l].windowOption );
					atom->support[0].len = 2*numFreqs[l];
					atom->support[0].pos = slotPos + frameShift[l] + m*numFreqs[l];
					atom->numSamples = atom->support[0].pos + atom->support[0].len;
					atom->freq = ((double)n+0.5)/(double)atom->support[0].len;
					atom->amp[0] = data[l][m][n];
					book->append(atom);
				}
			}
		}
	}
}

void slot::reset_data( void ) {

	/* Declarations */
	unsigned int l,m,n;

	/* Fill array with zeros */
	for ( l=0 ; l<numBlocks ; l++ ) {
		for ( m=0 ; m<numFrames[l] ; m++ ) {
			for ( n=0 ; n<numFreqs[l] ; n++ ) {
				data[l][m][n] = 0;
			}
		}
	}
}

void slot::interleave_data( void ) {

	/* Declarations */
	unsigned int block,frame,freq;

	/* Interleaving */
	for ( block=0 ; block<numBlocks ; block++ ) {
		for ( frame=0 ; frame<numFrames[block] ; frame++ ) {
			for ( freq=0 ; freq<numFreqs[block] ; freq++ ) {
				dataI[inter[block][frame][freq]] = data[block][frame][freq];
			}
		}
	}
}

void slot::deinterleave_data( void ) {

	/* Declarations */
	unsigned int block,frame,freq;

	/* Deinterleaving */
	for ( block=0 ; block<numBlocks ; block++ ) {
		for ( frame=0 ; frame<numFrames[block] ; frame++ ) {
			for ( freq=0 ; freq<numFreqs[block] ; freq++ ) {
				 data[block][frame][freq] = dataI[inter[block][frame][freq]];
			}
		}
	}
}

void slot::init_interleaving( void ) {

	/* Declarations */
	unsigned int l,m,n;
	unsigned int block;
	unsigned int framePermutOld[32768];
	unsigned int framePermutNew[32768];

	/* Interleaving */
	for ( n=0 ; n<numBlocks ; n++ ) {
		block = numBlocks-n-1;
		if (block==numBlocks-1) {
			framePermutOld[0] = 0;
		} else {
			for ( m=0 ; m<numFrames[block+1] ; m++ ) {
				framePermutNew[2*m] = framePermutOld[m];
				framePermutNew[2*m+1] = framePermutOld[m] + numFrames[block+1];
			}
			for ( m=0 ; m<numFrames[block] ; m++ ) {
				framePermutOld[m] = framePermutNew[m];
			}
		}
		for ( l=0 ; l<numFreqs[block] ; l++ ) {
			for ( m=0 ; m<numFrames[block] ; m++ ) {
				inter[block][framePermutOld[m]][l] = (l*numFrames[block]+m)*numBlocks+n;
			}
		}
	}
}

void slot::synthesize( void ) {

	/* Declarations */
	unsigned int block,frame,pos,len,n;

	/* Test if MDCT has been initialized */
	if ( mdct_blocks==NULL ) printf("Error: MDCT is not initialized in slot\n");

	/* Reset signal to zeros */
	for ( n=0 ; n<signal_len ; n++ ) signal[n] = 0;

	/* Inverse MDCT */
	for ( block=0 ; block<numBlocks ; block++ ) {
		len = 2*numFreqs[block];
		for ( frame=0 ; frame<numFrames[block] ; frame++ ) {
			pos = (numFreqs[numBlocks-1]>>1) + frameShift[block] + frame*numFreqs[block];
			mdct_blocks[block]->compute_inv_mdct( data[block][frame] );
			for ( n=0 ; n<len ; n++ ) signal[pos+n] += mdct_blocks[block]->samples[n];
		}
	}
}

void slot::compute_mask( slot *mask ) {

	/* Declarations */
	unsigned int block,frame,pos;

	/* Test if MDCT and PSY have been initialized */
	if ( mdct_blocks==NULL ) printf("Error: MDCT is not initialized in slot\n");
	if ( psy_blocks==NULL ) printf("Error: PSY model is not initialized in slot\n");

	/* Synthesize slot */
	synthesize();

	/* Compute mask threshold */
	for ( block=0 ; block<numBlocks ; block++ ) {
		for ( frame=0 ; frame<numFrames[block] ; frame++ ) {
			pos = (numFreqs[numBlocks-1]>>1) + frameShift[block] + frame*numFreqs[block];
			psy_blocks[block]->compute( mdct_blocks[block], signal+pos, mask->data[block][frame] );
		}
	}
}
