/******************************************************************************/
/*                                                                            */
/*                                 dict.cpp                                   */
/*                                                                            */
/*                        Matching Pursuit Library                            */
/*                                                                            */
/* R�mi Gribonval                                                             */
/* Sacha Krstulovic                                           Mon Feb 21 2005 */
/* -------------------------------------------------------------------------- */
/*                                                                            */
/*  Copyright (C) 2005 IRISA                                                  */
/*                                                                            */
/*  This program is free software; you can redistribute it and/or             */
/*  modify it under the terms of the GNU General Public License               */
/*  as published by the Free Software Foundation; either version 2            */
/*  of the License, or (at your option) any later version.                    */
/*                                                                            */
/*  This program is distributed in the hope that it will be useful,           */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/*  GNU General Public License for more details.                              */
/*                                                                            */
/*  You should have received a copy of the GNU General Public License         */
/*  along with this program; if not, write to the Free Software               */
/*  Foundation, Inc., 59 Temple Place - Suite 330,                            */
/*  Boston, MA  02111-1307, USA.                                              */
/*                                                                            */
/******************************************************************************/
/*
 * SVN log:
 *
 * $Author: sacha $
 * $Date: 2006-03-27 19:26:31 +0200 (lun, 27 mar 2006) $
 * $Revision: 548 $
 *
 */

/*****************************************/
/*                                       */
/* dict.cpp: methods for class MP_Dict_c */
/*                                       */
/*****************************************/

#include "mptk.h"
#include "mp_system.h"


/***************************/
/* CONSTRUCTORS/DESTRUCTOR */
/***************************/


/*************************************************/
/* Plain initialization, with no data whatsoever */
MP_Dict_c* MP_Dict_c::init( void ) {

  const char* func = "MP_Dict_c::init(void)";
  MP_Dict_c *newDict = NULL;

  /* Instantiate and check */
  newDict = new MP_Dict_c();
  if ( newDict == NULL ) {
    mp_error_msg( func, "Failed to create a new dictionary.\n" );
    return( NULL );
  }

  return( newDict );
}


/**************/
/* NULL constructor */
MP_Dict_c::MP_Dict_c() {

  mp_debug_msg( MP_DEBUG_CONSTRUCTION, "MP_Dict_c::MP_Dict_c()",
		"Constructing dict...\n" );

  signal = NULL;
  sigMode = MP_DICT_NULL_SIGNAL;
  numBlocks = 0;
  block = NULL;
  blockWithMaxIP = UINT_MAX;
  touch = NULL;

  mp_debug_msg( MP_DEBUG_CONSTRUCTION, "MP_Dict_c::MP_Dict_c()",
		"Done.\n" );
}


/**************/
/* Destructor */
MP_Dict_c::~MP_Dict_c() {

  unsigned int i;

  mp_debug_msg( MP_DEBUG_DESTRUCTION, "MP_Dict_c::~MP_Dict_c()",
		"Deleting dict...\n" );

  if ( (sigMode == MP_DICT_INTERNAL_SIGNAL) && ( signal != NULL ) ) delete signal;
  if ( block ) {
    for ( i=0; i<numBlocks; i++ ) { if ( block[i] ) delete( block[i] ); }
    free( block );
  }
  if ( touch ) free( touch );

  mp_debug_msg( MP_DEBUG_DESTRUCTION, "MP_Dict_c::~MP_Dict_c()",
		"Done.\n" );
}



/***************************/
/* I/O METHODS             */
/***************************/

/******************************/
/* Addition of a single block */
int MP_Dict_c::add_block( MP_Block_c *newBlock ) {

  const char* func = "MP_Dict_c::add_block( newBlock )";
  MP_Block_c **tmp;

  if( newBlock != NULL ) {

    /* Increase the size of the array of blocks... */
    if ( (tmp = (MP_Block_c**) realloc( block, (numBlocks+1)*sizeof(MP_Block_c*) )) == NULL ) {
      mp_error_msg( func, "Can't reallocate memory to add a new block to dictionary."
		    " No new block will be added, number of blocks stays [%d].\n",
		    numBlocks );
      return( 0 );
    }
    /* ... and store the reference on the newly created object. */
    else {
      block = tmp;
      block[numBlocks] = newBlock;
      numBlocks++;
    }
    return( 1 );

  }
  /* else, if the block is NULL, silently ignore it,
     just say that no block has been added. */
  else return( 0 );

}

/***************************/
/* MISC METHODS            */
/***************************/

/******************************/
/* Return the number of atoms */
unsigned long int MP_Dict_c::num_atoms(void) {

  unsigned long int numAtoms = 0;
  unsigned int i;

  for (i = 0; i < numBlocks; i++)
    numAtoms += block[i]->num_atoms();

  return(numAtoms);
}


/*****************************************/
/* Copy a new signal into the dictionary */
int MP_Dict_c::copy_signal( MP_Signal_c *setSignal ) {

  const char* func = "MP_Dict_c::copy_signal( signal )";
  unsigned int i;
  int check = 0;

  if ( signal ) { delete( signal ); signal = NULL; }
  if ( setSignal != NULL ) {
    signal = new MP_Signal_c( *setSignal );
    sigMode = MP_DICT_INTERNAL_SIGNAL;
  }
  else {
    signal = NULL;
    sigMode = MP_DICT_NULL_SIGNAL;
  }

  /* Allocate the touch array
     (alloc_touch() will automatically manage the NULL signal case) */
  if ( alloc_touch() ) {
    mp_error_msg( func, "Failed to allocate and initialize the touch array"
		  " in the dictionary constructor. Signal and touch will stay NULL.\n" );
    delete( signal );
    signal = NULL;
    sigMode = MP_DICT_NULL_SIGNAL;
    return( 1 );
  }

  /* Then plug the signal into all the blocks */
  for ( i = 0; i < numBlocks; i++ ) {
    if ( block[i]->plug_signal( signal ) ) {
      mp_error_msg( func, "Failed to plug the given signal in the %u-th block."
		    " Proceeding with the remaining blocks.\n", i );
      check = 1;
    }
  }

  return( check );
}


/*****************************************/
/* Load a new signal into the dictionary,
   from a file */
int MP_Dict_c::copy_signal( const char *fName ) {

  const char* func = "MP_Dict_c::copy_signal( fileName )";
  unsigned int i;
  int check = 0;

  if( fName == NULL ) {
    mp_error_msg( func, "Passed a NULL string for the file name.\n" );
    return( 1 );
  }

  if ( signal ) { delete( signal ); signal = NULL; }
  signal = MP_Signal_c::init( fName );
  if ( signal == NULL ) {
    mp_error_msg( func, "Failed to instantiate a signal from file [%s]\n",
		  fName );
    sigMode = MP_DICT_NULL_SIGNAL;
    alloc_touch(); /* -> Nullifies the touch array when signal is NULL. */
    return( 1 );
  }
  /* else  */
  sigMode = MP_DICT_INTERNAL_SIGNAL;

  /* Allocate the touch array */
  if ( alloc_touch() ) {
    mp_error_msg( func, "Failed to allocate and initialize the touch array"
		  " in the dictionary constructor. Signal and touch will stay NULL.\n" );
    delete( signal );
    signal = NULL;
    sigMode = MP_DICT_NULL_SIGNAL;
    return( 1 );
  }

  /* Then plug the signal into all the blocks */
  for ( i = 0; i < numBlocks; i++ ) {
    if ( block[i]->plug_signal( signal ) ) {
      mp_error_msg( func, "Failed to plug the given signal in the %u-th block."
		    " Proceeding with the remaining blocks.\n", i );
      check = 1;
    }
  }

  return( check );
}


/*****************************************/
/* Copy a new signal into the dictionary */
int MP_Dict_c::plug_signal( MP_Signal_c *setSignal ) {

  const char* func = "MP_Dict_c::plug_signal( signal )";
  int check = 0;
  unsigned int i;

  if ( setSignal != NULL ) {
    signal = setSignal;
    sigMode = MP_DICT_EXTERNAL_SIGNAL;
  }
  else {
    signal = NULL;
    sigMode = MP_DICT_NULL_SIGNAL;
  }

  /* Allocate the touch array
     (alloc_touch() will automatically manage the NULL signal case) */
  if ( alloc_touch() ) {
    mp_error_msg( func, "Failed to allocate and initialize the touch array"
		  " in the dictionary constructor. Signal and touch will stay NULL.\n" );
    signal = NULL;
    sigMode = MP_DICT_NULL_SIGNAL;
    return( 1 );
  }

  /* Then plug the signal into all the blocks */
  for ( i = 0; i < numBlocks; i++ ) {
    if ( block[i]->plug_signal( signal ) ) {
      mp_error_msg( func, "Failed to plug the given signal in the %u-th block."
		    " Proceeding with the remaining blocks.\n", i );
      check = 1;
    }
  }

  return( check );
}


/*****************************************/
/* Detach the signal from the dictionary */
MP_Signal_c* MP_Dict_c::detach_signal( void ) {

  MP_Signal_c *s = signal;

  plug_signal( NULL );

  return( s );
}


/*********************************/
/* Allocation of the touch array */
int MP_Dict_c::alloc_touch( void ) {

  const char* func = "MP_Dict_c::alloc_touch()";

  /* If touch already exists, free it; guarantee that it is NULL. */
  if ( touch ) { free( touch ); touch = NULL; }

  /* Check if a signal is present: */
  if ( signal != NULL ) {

    /* Allocate the touch array */
    if ( (touch = (MP_Support_t*) malloc( signal->numChans*sizeof(MP_Support_t) )) == NULL ) {
      mp_error_msg( func, "Failed to allocate the array of touched supports"
		    " in the dictionary. The touch array is left NULL.\n" );
      return( 1 );
    }
    /* If the allocation went OK, initialize the array:*/
    else {
      int i;
      /* Initially we have to consider the whole signal as touched */
      for ( i = 0; i < signal->numChans; i++ ) { touch[i].pos = 0; touch[i].len = signal->numSamples; }
      return( 0 );
    }

  }
  else {
    mp_error_msg( func, "Trying to allocate the touch array from"
		  " a NULL signal. The touch array will remain NULL.\n" );
    return( 1 );
  }

}


/*************************/
/* Delete all the blocks */
int MP_Dict_c::delete_all_blocks( void ) {

  unsigned int i;
  unsigned int oldNumBlocks = numBlocks;

  for( i=0; i<numBlocks; i++) {
    if ( block[i] ) delete( block[i] );
  }
  if ( block ) free( block );
  block = NULL;
  numBlocks = 0;
  blockWithMaxIP = 0;

  return( oldNumBlocks );
}


/************************************/
/* Update of all the inner products which need to be updated, according to the 'touch' field */
MP_Real_t MP_Dict_c::update( void ) {

  unsigned int i;
  MP_Real_t tempMax = -1.0;
  MP_Real_t val;
  MP_Support_t frameSupport;

  for( i=0; i<numBlocks; i++) {
    /* Recompute the inner products */
    frameSupport = block[i]->update_ip( touch );
    /* Recompute and locate the max inner product within a block */
    val = block[i]->update_max( frameSupport );
    /* Locate the max inner product across the blocks */
    if ( val > tempMax ) { tempMax = val; blockWithMaxIP = i; }
  }

  return ( tempMax );
}

/**********************************************/
/* Forces an update of all the inner products */
MP_Real_t MP_Dict_c::update_all( void ) {

  unsigned int i;
  MP_Real_t tempMax = -1.0;
  MP_Real_t val;
  MP_Support_t frameSupport;

  for( i=0; i<numBlocks; i++) {
    /* (Re)compute all the inner products */
    frameSupport = block[i]->update_ip( NULL );
    /* Refresh and locate the max inner product within a block */
    val = block[i]->update_max( frameSupport );
    /* Locate the max inner product across the blocks */
    if ( val > tempMax ) { tempMax = val; blockWithMaxIP = i; }
  }
  return( tempMax );
}

/************************************/
/* Create a new atom corresponding to the best atom of the best block. */
unsigned int MP_Dict_c::create_max_atom( MP_Atom_c** atom ) {

  const char* func = "MP_Dict_c::create_max_atom(**atom)";
  unsigned long int frameIdx;
  unsigned long int filterIdx;
  unsigned int numAtoms;
  int chanIdx;
  MP_Support_t frameSupport;
  MP_Real_t tempMax,val;

  /* 1) select the best atom of the best block */
  frameIdx =  block[blockWithMaxIP]->maxIPFrameIdx;
  filterIdx = block[blockWithMaxIP]->maxIPIdxInFrame[frameIdx];

  mp_debug_msg( MP_DEBUG_MP_ITERATIONS, func,
		"Found max atom in block [%lu], at frame [%lu], with filterIdx [%lu].\n",
		blockWithMaxIP, frameIdx, filterIdx );

  /* 2) create it using the method of the block */
  numAtoms = block[blockWithMaxIP]->create_atom( atom, frameIdx, filterIdx );
  if ( numAtoms == 0 ) {
    mp_error_msg( func, "Failed to create the max atom from block[%ul].\n",
		  blockWithMaxIP );
    return( 0 );
  }

  /* 3) Optionnal: pre-echo control */
  if (block[blockWithMaxIP]->pec&&block[blockWithMaxIP]->test_preEcho(*atom)) {
    for ( chanIdx=0; chanIdx < (*atom)->numChans; chanIdx++ ) {
      touch[chanIdx].pos = (*atom)->support[chanIdx].pos;
      touch[chanIdx].len = (*atom)->support[chanIdx].len;
    }
    frameSupport = block[blockWithMaxIP]->update_ip( touch );
    tempMax = block[blockWithMaxIP]->update_max( frameSupport );
    for( unsigned int i=0; i<numBlocks; i++) {
      val = block[i]->maxIPValue;
      if ( val > tempMax ) { tempMax = val; blockWithMaxIP = i; }
    }
    numAtoms = create_max_atom( atom );
  }

  return( numAtoms );
}

/******************************************/
/* Perform one matching pursuit iteration */
int MP_Dict_c::iterate_mp( MP_Book_c *book , MP_Signal_c *sigRecons ) {

  const char* func = "MP_Dict_c::iterate_mp(...)";
  int chanIdx;
  MP_Atom_c *atom;
  unsigned int numAtoms;

  /* Check if a signal is present */
  if ( signal == NULL ) {
    mp_error_msg( func, "There is no signal in the dictionary. You must"
		  " plug or copy a signal before you can iterate.\n" );
    return( 1 );
  }

  /* 1/ refresh the inner products
   * 2/ create the max atom and store it
   * 3/ substract it from the signal and add it to the approximant
   */

  /** 1/ (Re)compute the inner products according to the current 'touch' indicating where the signal 
   * may have been changed after the last update of the inner products */
  update();

  /** 2/ Create the max atom and store it in the book */
  numAtoms = create_max_atom( &atom );
  if ( numAtoms == 0 ) {
    mp_error_msg( func, "The Matching Pursuit iteration failed. Dictionary, book"
		  " and signal are left unchanged.\n" );
    return( 1 );
  }
  
  if ( book->append( atom ) != 1 ) {
    mp_error_msg( func, "Failed to append the max atom to the book.\n" );
    return( 1 );
  }

  /* 3/ Substract the atom's waveform from the analyzed signal */
  atom->substract_add( signal , sigRecons );

  /* 4/ Keep track of the support where the signal has been modified */
  for ( chanIdx=0; chanIdx < atom->numChans; chanIdx++ ) {
    touch[chanIdx].pos = atom->support[chanIdx].pos;
    touch[chanIdx].len = atom->support[chanIdx].len;
  } 

  return( 0 );
}
